# automatically generated by the FlatBuffers compiler, do not modify

# namespace: reflection

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class RPCCall(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RPCCall()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsRPCCall(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    @classmethod
    def RPCCallBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x42\x46\x42\x53", size_prefixed=size_prefixed
        )

    # RPCCall
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RPCCall
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RPCCall
    def Request(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from zlmdb.flatbuffers.reflection.Object import Object

            obj = Object()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RPCCall
    def Response(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from zlmdb.flatbuffers.reflection.Object import Object

            obj = Object()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RPCCall
    def Attributes(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from zlmdb.flatbuffers.reflection.KeyValue import KeyValue

            obj = KeyValue()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RPCCall
    def AttributesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # RPCCall
    def AttributesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # RPCCall
    def Documentation(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4)
            )
        return ""

    # RPCCall
    def DocumentationLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # RPCCall
    def DocumentationIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0


def RPCCallStart(builder):
    builder.StartObject(5)


def Start(builder):
    return RPCCallStart(builder)


def RPCCallAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0
    )


def AddName(builder, name):
    return RPCCallAddName(builder, name)


def RPCCallAddRequest(builder, request):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(request), 0
    )


def AddRequest(builder, request):
    return RPCCallAddRequest(builder, request)


def RPCCallAddResponse(builder, response):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(response), 0
    )


def AddResponse(builder, response):
    return RPCCallAddResponse(builder, response)


def RPCCallAddAttributes(builder, attributes):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(attributes), 0
    )


def AddAttributes(builder, attributes):
    return RPCCallAddAttributes(builder, attributes)


def RPCCallStartAttributesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartAttributesVector(builder, numElems):
    return RPCCallStartAttributesVector(builder, numElems)


def RPCCallAddDocumentation(builder, documentation):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(documentation), 0
    )


def AddDocumentation(builder, documentation):
    return RPCCallAddDocumentation(builder, documentation)


def RPCCallStartDocumentationVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartDocumentationVector(builder, numElems):
    return RPCCallStartDocumentationVector(builder, numElems)


def RPCCallEnd(builder):
    return builder.EndObject()


def End(builder):
    return RPCCallEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import pprint
import uuid

import numpy as np

import flatbuffers

from zlmdb import table, MapTimestampUuidFlatBuffers
from txaio import time_ns

import MNodeLog as MNodeLogGen


class _MNodeLogGen(MNodeLogGen.MNodeLog):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsMNodeLog(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _MNodeLogGen()
        x.Init(buf, n + offset)
        return x

    def NodeIdAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def RunIdAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class MNodeLog(object):
    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # uint64
        self._timestamp = None

        # node_id: [uint8] (uuid);
        self._node_id = None

        # run_id: [uint8] (uuid);
        self._run_id = None

        # state: MNodeState;
        self._state = None

        # ended: uint64;
        self._ended = None

        # session: uint64;
        self._session = None

        # sent: uint64;
        self._sent = None

        # seq: uint64;
        self._seq = None

        # routers: uint16;
        self._routers = None

        # containers: uint16;
        self._containers = None

        # guests: uint16;
        self._guests = None

        # proxies: uint16;
        self._proxies = None

        # xbr_marketmakers: uint16;
        self._marketmakers = None

        # CF node system statistics
        self._cpu_ctx_switches = None
        self._cpu_freq = None
        self._cpu_guest = None
        self._cpu_guest_nice = None
        self._cpu_idle = None
        self._cpu_interrupts = None
        self._cpu_iotwait = None
        self._cpu_irq = None
        self._cpu_nice = None
        self._cpu_soft_interrupts = None
        self._cpu_softirq = None
        self._cpu_steal = None
        self._cpu_system = None
        self._cpu_user = None
        self._disk_busy_time = None
        self._disk_read_bytes = None
        self._disk_read_count = None
        self._disk_read_merged_count = None
        self._disk_read_time = None
        self._disk_write_bytes = None
        self._disk_write_count = None
        self._disk_write_merged_count = None
        self._disk_write_time = None
        self._memory_active = None
        self._memory_available = None
        self._memory_buffers = None
        self._memory_cached = None
        self._memory_free = None
        self._memory_inactive = None
        self._memory_percent = None
        self._memory_shared = None
        self._memory_slab = None
        self._memory_total = None
        self._memory_used = None
        self._network_bytes_recv = None
        self._network_bytes_sent = None
        self._network_connection_af_inet = None
        self._network_connection_af_inet6 = None
        self._network_connection_af_unix = None
        self._network_dropin = None
        self._network_dropout = None
        self._network_errin = None
        self._network_errout = None
        self._network_packets_recv = None
        self._network_packets_sent = None

    @staticmethod
    def parse(node_id, heartbeat):
        assert isinstance(node_id, uuid.UUID)
        assert type(heartbeat) == dict
        assert "timestamp" in heartbeat and type(heartbeat["timestamp"]) == int

        obj = MNodeLog()
        obj._timestamp = np.datetime64(time_ns(), "ns")
        obj._node_id = node_id
        obj._run_id = uuid.UUID(bytes=b"\0" * 16)
        obj._state = heartbeat.get("state", None)
        obj._ended = (
            np.datetime64(heartbeat["ended"], "ns")
            if heartbeat.get("ended", None)
            else None
        )
        obj._session = heartbeat.get("session", None)
        obj._sent = (
            np.datetime64(heartbeat["timestamp"], "ns")
            if heartbeat.get("timestamp", None)
            else None
        )
        obj._seq = heartbeat.get("seq", None)

        workers = heartbeat.get("workers", {})
        obj._routers = workers.get("router", None)
        obj._containers = workers.get("container", None)
        obj._guests = workers.get("guest", None)
        obj._proxies = workers.get("proxy", None)
        obj._marketmakers = workers.get("xbrmm", None)

        system = heartbeat.get("system", {})
        system_cpu = system.get("cpu", {})
        system_net = system.get("network", {})
        system_mem = system.get("memory", {})
        system_dsk = system.get("disk", {})

        obj._cpu_ctx_switches = system_cpu.get("ctx_switches", None)
        obj._cpu_freq = system_cpu.get("freq", None)
        obj._cpu_guest = system_cpu.get("guest", None)
        obj._cpu_guest_nice = system_cpu.get("guest_nice", None)
        obj._cpu_idle = system_cpu.get("idle", None)
        obj._cpu_interrupts = system_cpu.get("interrupts", None)
        obj._cpu_iotwait = system_cpu.get("iotwait", None)
        obj._cpu_irq = system_cpu.get("irq", None)
        obj._cpu_nice = system_cpu.get("nice", None)
        obj._cpu_soft_interrupts = system_cpu.get("soft_interrupts", None)
        obj._cpu_softirq = system_cpu.get("softirq", None)
        obj._cpu_steal = system_cpu.get("steal", None)
        obj._cpu_system = system_cpu.get("system", None)
        obj._cpu_user = system_cpu.get("user", None)

        obj._network_bytes_recv = system_net.get("bytes_recv", None)
        obj._network_bytes_sent = system_net.get("bytes_sent", None)
        obj._network_packets_recv = system_net.get("packets_recv", None)
        obj._network_packets_sent = system_net.get("packets_sent", None)
        obj._network_dropin = system_net.get("dropin", None)
        obj._network_dropout = system_net.get("dropout", None)
        obj._network_errin = system_net.get("errin", None)
        obj._network_errout = system_net.get("errout", None)

        connection = system_net.get("connection", {})
        obj._network_connection_af_inet = connection.get("AF_INET", None)
        obj._network_connection_af_inet6 = connection.get("AF_INET6", None)
        obj._network_connection_af_unix = connection.get("AF_UNIX", None)

        obj._memory_active = system_mem.get("active", None)
        obj._memory_available = system_mem.get("available", None)
        obj._memory_buffers = system_mem.get("buffers", None)
        obj._memory_cached = system_mem.get("cached", None)
        obj._memory_free = system_mem.get("free", None)
        obj._memory_inactive = system_mem.get("inactive", None)
        obj._memory_percent = system_mem.get("percent", None)
        obj._memory_shared = system_mem.get("shared", None)
        obj._memory_slab = system_mem.get("slab", None)
        obj._memory_total = system_mem.get("total", None)
        obj._memory_used = system_mem.get("used", None)

        obj._disk_busy_time = system_dsk.get("busy_time", None)
        obj._disk_read_bytes = system_dsk.get("read_bytes", None)
        obj._disk_read_count = system_dsk.get("read_count", None)
        obj._disk_read_merged_count = system_dsk.get("read_merged_count", None)
        obj._disk_read_time = system_dsk.get("read_time", None)
        obj._disk_write_bytes = system_dsk.get("write_bytes", None)
        obj._disk_write_count = system_dsk.get("write_count", None)
        obj._disk_write_merged_count = system_dsk.get("write_merged_count", None)
        obj._disk_write_time = system_dsk.get("write_time", None)

        return obj

    def marshal(self):
        obj = {
            "timestamp": self.timestamp,
            "node_id": str(self.node_id),
            "run_id": str(self.run_id),
            "state": self.state,
            "ended": self.ended,
            "session": self.session,
            "sent": self.sent,
            "seq": self.seq,
            "workers": {
                "router": self.routers,
                "container": self.containers,
                "guest": self.guests,
                "proxy": self.proxies,
                "xbrmm": self.marketmakers,
            },
            "cpu": {
                "ctx_switches": self.cpu_ctx_switches,
                "freq": self.cpu_freq,
                "guest": self.cpu_guest,
                "guest_nice": self.cpu_guest_nice,
                "idle": self.cpu_idle,
                "interrupts": self.cpu_interrupts,
                "iotwait": self.cpu_iotwait,
                "irq": self.cpu_irq,
                "nice": self.cpu_nice,
                "soft_interrupts": self.cpu_soft_interrupts,
                "softirq": self.cpu_softirq,
                "steal": self.cpu_steal,
                "system": self.cpu_system,
                "user": self.cpu_user,
            },
            "memory": {
                "active": self.memory_active,
                "available": self.memory_available,
                "buffers": self.memory_buffers,
                "cached": self.memory_cached,
                "free": self.memory_free,
                "inactive": self.memory_inactive,
                "percent": self.memory_percent,
                "shared": self.memory_shared,
                "slab": self.memory_slab,
                "total": self.memory_total,
                "used": self.memory_used,
            },
            "disk": {
                "busy_time": self.disk_busy_time,
                "read_bytes": self.disk_read_bytes,
                "read_count": self.disk_read_count,
                "read_merged_count": self.disk_read_merged_count,
                "read_time": self.disk_read_time,
                "write_bytes": self.disk_write_bytes,
                "write_count": self.disk_write_count,
                "write_merged_count": self.disk_write_merged_count,
                "write_time": self.disk_write_time,
            },
            "network": {
                "bytes_recv": self.network_bytes_recv,
                "bytes_sent": self.network_bytes_sent,
                "connection": {
                    "AF_INET": self.network_connection_af_inet,
                    "AF_INET6": self.network_connection_af_inet6,
                    "UNIX": self.network_connection_af_unix,
                },
                "dropin": self.network_dropin,
                "dropout": self.network_dropout,
                "errin": self.network_errin,
                "errout": self.network_errout,
                "packets_recv": self.network_packets_recv,
                "packets_sent": self.network_packets_sent,
            },
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def timestamp(self):
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def node_id(self):
        if self._node_id is None and self._from_fbs:
            if self._from_fbs.NodeIdLength():
                _node_id = self._from_fbs.NodeIdAsBytes()
                self._node_id = uuid.UUID(bytes=_node_id.tobytes())
        return self._node_id

    @node_id.setter
    def node_id(self, value):
        assert value is None or isinstance(value, uuid.UUID)
        self._node_id = value

    @property
    def run_id(self):
        if self._run_id is None and self._from_fbs:
            if self._from_fbs.RunIdLength():
                _run_id = self._from_fbs.RunIdAsBytes()
                self._run_id = uuid.UUID(bytes=_run_id.tobytes())
        return self._run_id

    @run_id.setter
    def run_id(self, value):
        assert value is None or isinstance(value, uuid.UUID)
        self._run_id = value

    @property
    def state(self):
        if self._state is None and self._from_fbs:
            self._state = self._from_fbs.State()
        return self._state

    @state.setter
    def state(self, value):
        assert value is None or type(value) == int
        self._state = value

    @property
    def ended(self):
        if self._ended is None and self._from_fbs:
            self._ended = np.datetime64(self._from_fbs.Ended(), "ns")
        return self._ended

    @ended.setter
    def ended(self, value):
        assert value is None or isinstance(value, np.datetime64)
        self._ended = value

    @property
    def session(self):
        if self._session is None and self._from_fbs:
            self._session = self._from_fbs.Session()
        return self._session

    @session.setter
    def session(self, value):
        assert value is None or type(value) == int
        self._session = value

    @property
    def sent(self):
        if self._sent is None and self._from_fbs:
            self._sent = np.datetime64(self._from_fbs.Sent(), "ns")
        return self._sent

    @sent.setter
    def sent(self, value):
        assert value is None or isinstance(value, np.datetime64)
        self._sent = value

    @property
    def seq(self):
        if self._seq is None and self._from_fbs:
            self._seq = self._from_fbs.Seq()
        return self._seq

    @seq.setter
    def seq(self, value):
        assert value is None or type(value) == int
        self._seq = value

    @property
    def routers(self):
        if self._routers is None and self._from_fbs:
            self._routers = self._from_fbs.Routers()
        return self._routers

    @routers.setter
    def routers(self, value):
        assert value is None or type(value) == int
        self._routers = value

    @property
    def containers(self):
        if self._containers is None and self._from_fbs:
            self._containers = self._from_fbs.Containers()
        return self._containers

    @containers.setter
    def containers(self, value):
        assert value is None or type(value) == int
        self._containers = value

    @property
    def guests(self):
        if self._guests is None and self._from_fbs:
            self._guests = self._from_fbs.Guests()
        return self._guests

    @guests.setter
    def guests(self, value):
        assert value is None or type(value) == int
        self._guests = value

    @property
    def proxies(self):
        if self._proxies is None and self._from_fbs:
            self._proxies = self._from_fbs.Proxies()
        return self._proxies

    @proxies.setter
    def proxies(self, value):
        assert value is None or type(value) == int
        self._proxies = value

    @property
    def marketmakers(self):
        if self._marketmakers is None and self._from_fbs:
            self._marketmakers = self._from_fbs.Marketmakers()
        return self._marketmakers

    @marketmakers.setter
    def marketmakers(self, value):
        assert value is None or type(value) == int
        self._marketmakers = value

    @property
    def cpu_ctx_switches(self):
        if self._cpu_ctx_switches is None and self._from_fbs:
            self._cpu_ctx_switches = self._from_fbs.CpuCtxSwitches()
        return self._cpu_ctx_switches

    @cpu_ctx_switches.setter
    def cpu_ctx_switches(self, value):
        assert value is None or type(value) == int
        self._cpu_ctx_switches = value

    @property
    def cpu_freq(self):
        if self._cpu_freq is None and self._from_fbs:
            self._cpu_freq = self._from_fbs.CpuFreq()
        return self._cpu_freq

    @cpu_freq.setter
    def cpu_freq(self, value):
        assert value is None or type(value) == float
        self._cpu_freq = value

    @property
    def cpu_guest(self):
        if self._cpu_guest is None and self._from_fbs:
            self._cpu_guest = self._from_fbs.CpuGuest()
        return self._cpu_guest

    @cpu_guest.setter
    def cpu_guest(self, value):
        assert value is None or type(value) == float
        self._cpu_guest = value

    @property
    def cpu_guest_nice(self):
        if self._cpu_guest_nice is None and self._from_fbs:
            self._cpu_guest_nice = self._from_fbs.CpuGuestNice()
        return self._cpu_guest_nice

    @cpu_guest_nice.setter
    def cpu_guest_nice(self, value):
        assert value is None or type(value) == float
        self._cpu_guest_nice = value

    @property
    def cpu_idle(self):
        if self._cpu_idle is None and self._from_fbs:
            self._cpu_idle = self._from_fbs.CpuIdle()
        return self._cpu_idle

    @cpu_idle.setter
    def cpu_idle(self, value):
        assert value is None or type(value) == float
        self._cpu_idle = value

    @property
    def cpu_interrupts(self):
        if self._cpu_interrupts is None and self._from_fbs:
            self._cpu_interrupts = self._from_fbs.CpuInterrupts()
        return self._cpu_interrupts

    @cpu_interrupts.setter
    def cpu_interrupts(self, value):
        assert value is None or type(value) == int
        self._cpu_interrupts = value

    @property
    def cpu_iotwait(self):
        if self._cpu_iotwait is None and self._from_fbs:
            self._cpu_iotwait = self._from_fbs.CpuIowait()
        return self._cpu_iotwait

    @cpu_iotwait.setter
    def cpu_iotwait(self, value):
        assert value is None or type(value) == float
        self._cpu_iotwait = value

    @property
    def cpu_irq(self):
        if self._cpu_irq is None and self._from_fbs:
            self._cpu_irq = self._from_fbs.CpuIrq()
        return self._cpu_irq

    @cpu_irq.setter
    def cpu_irq(self, value):
        assert value is None or type(value) == float
        self._cpu_irq = value

    @property
    def cpu_nice(self):
        if self._cpu_nice is None and self._from_fbs:
            self._cpu_nice = self._from_fbs.CpuNice()
        return self._cpu_nice

    @cpu_nice.setter
    def cpu_nice(self, value):
        assert value is None or type(value) == float
        self._cpu_nice = value

    @property
    def cpu_soft_interrupts(self):
        if self._cpu_soft_interrupts is None and self._from_fbs:
            self._cpu_soft_interrupts = self._from_fbs.CpuSoftInterrupts()
        return self._cpu_soft_interrupts

    @cpu_soft_interrupts.setter
    def cpu_soft_interrupts(self, value):
        assert value is None or type(value) == int
        self._cpu_soft_interrupts = value

    @property
    def cpu_softirq(self):
        if self._cpu_softirq is None and self._from_fbs:
            self._cpu_softirq = self._from_fbs.CpuSoftirq()
        return self._cpu_softirq

    @cpu_softirq.setter
    def cpu_softirq(self, value):
        assert value is None or type(value) == float
        self._cpu_softirq = value

    @property
    def cpu_steal(self):
        if self._cpu_steal is None and self._from_fbs:
            self._cpu_steal = self._from_fbs.CpuSteal()
        return self._cpu_steal

    @cpu_steal.setter
    def cpu_steal(self, value):
        assert value is None or type(value) == float
        self._cpu_steal = value

    @property
    def cpu_system(self):
        if self._cpu_system is None and self._from_fbs:
            self._cpu_system = self._from_fbs.CpuSystem()
        return self._cpu_system

    @cpu_system.setter
    def cpu_system(self, value):
        assert value is None or type(value) == float
        self._cpu_system = value

    @property
    def cpu_user(self):
        if self._cpu_user is None and self._from_fbs:
            self._cpu_user = self._from_fbs.CpuUser()
        return self._cpu_user

    @cpu_user.setter
    def cpu_user(self, value):
        assert value is None or type(value) == float
        self._cpu_user = value

    @property
    def network_bytes_recv(self):
        if self._network_bytes_recv is None and self._from_fbs:
            self._network_bytes_recv = self._from_fbs.NetworkBytesRecv()
        return self._network_bytes_recv

    @network_bytes_recv.setter
    def network_bytes_recv(self, value):
        assert value is None or type(value) == int
        self._network_bytes_recv = value

    @property
    def network_bytes_sent(self):
        if self._network_bytes_sent is None and self._from_fbs:
            self._network_bytes_sent = self._from_fbs.NetworkBytesSent()
        return self._network_bytes_sent

    @network_bytes_sent.setter
    def network_bytes_sent(self, value):
        assert value is None or type(value) == int
        self._network_bytes_sent = value

    @property
    def network_connection_af_inet(self):
        if self._network_connection_af_inet is None and self._from_fbs:
            self._network_connection_af_inet = self._from_fbs.NetworkConnectionAfInet()
        return self._network_connection_af_inet

    @network_connection_af_inet.setter
    def network_connection_af_inet(self, value):
        assert value is None or type(value) == int
        self._network_connection_af_inet = value

    @property
    def network_connection_af_inet6(self):
        if self._network_connection_af_inet6 is None and self._from_fbs:
            self._network_connection_af_inet6 = (
                self._from_fbs.NetworkConnectionAfInet6()
            )
        return self._network_connection_af_inet6

    @network_connection_af_inet6.setter
    def network_connection_af_inet6(self, value):
        assert value is None or type(value) == int
        self._network_connection_af_inet6 = value

    @property
    def network_connection_af_unix(self):
        if self._network_connection_af_unix is None and self._from_fbs:
            self._network_connection_af_unix = self._from_fbs.NetworkConnectionAfUnix()
        return self._network_connection_af_unix

    @network_connection_af_unix.setter
    def network_connection_af_unix(self, value):
        assert value is None or type(value) == int
        self._network_connection_af_unix = value

    @property
    def network_dropin(self):
        if self._network_dropin is None and self._from_fbs:
            self._network_dropin = self._from_fbs.NetworkDropin()
        return self._network_dropin

    @network_dropin.setter
    def network_dropin(self, value):
        assert value is None or type(value) == int
        self._network_dropin = value

    @property
    def network_dropout(self):
        if self._network_dropout is None and self._from_fbs:
            self._network_dropout = self._from_fbs.NetworkDropout()
        return self._network_dropout

    @network_dropout.setter
    def network_dropout(self, value):
        assert value is None or type(value) == int
        self._network_dropout = value

    @property
    def network_errin(self):
        if self._network_errin is None and self._from_fbs:
            self._network_errin = self._from_fbs.NetworkErrin()
        return self._network_errin

    @network_errin.setter
    def network_errin(self, value):
        assert value is None or type(value) == int
        self._network_errin = value

    @property
    def network_errout(self):
        if self._network_errout is None and self._from_fbs:
            self._network_errout = self._from_fbs.NetworkErrout()
        return self._network_errout

    @network_errout.setter
    def network_errout(self, value):
        assert value is None or type(value) == int
        self._network_errout = value

    @property
    def network_packets_recv(self):
        if self._network_packets_recv is None and self._from_fbs:
            self._network_packets_recv = self._from_fbs.NetworkPacketsRecv()
        return self._network_packets_recv

    @network_packets_recv.setter
    def network_packets_recv(self, value):
        assert value is None or type(value) == int
        self._network_packets_recv = value

    @property
    def network_packets_sent(self):
        if self._network_packets_sent is None and self._from_fbs:
            self._network_packets_sent = self._from_fbs.NetworkPacketsSent()
        return self._network_packets_sent

    @network_packets_sent.setter
    def network_packets_sent(self, value):
        assert value is None or type(value) == int
        self._network_packets_sent = value

    @property
    def memory_active(self):
        if self._memory_active is None and self._from_fbs:
            self._memory_active = self._from_fbs.MemoryActive()
        return self._memory_active

    @memory_active.setter
    def memory_active(self, value):
        assert value is None or type(value) == int
        self._memory_active = value

    @property
    def memory_available(self):
        if self._memory_available is None and self._from_fbs:
            self._memory_available = self._from_fbs.MemoryAvailable()
        return self._memory_available

    @memory_available.setter
    def memory_available(self, value):
        assert value is None or type(value) == int
        self._memory_available = value

    @property
    def memory_buffers(self):
        if self._memory_buffers is None and self._from_fbs:
            self._memory_buffers = self._from_fbs.MemoryBuffers()
        return self._memory_buffers

    @memory_buffers.setter
    def memory_buffers(self, value):
        assert value is None or type(value) == int
        self._memory_buffers = value

    @property
    def memory_cached(self):
        if self._memory_cached is None and self._from_fbs:
            self._memory_cached = self._from_fbs.MemoryCached()
        return self._memory_cached

    @memory_cached.setter
    def memory_cached(self, value):
        assert value is None or type(value) == int
        self._memory_cached = value

    @property
    def memory_free(self):
        if self._memory_free is None and self._from_fbs:
            self._memory_free = self._from_fbs.MemoryFree()
        return self._memory_free

    @memory_free.setter
    def memory_free(self, value):
        assert value is None or type(value) == int
        self._memory_free = value

    @property
    def memory_inactive(self):
        if self._memory_inactive is None and self._from_fbs:
            self._memory_inactive = self._from_fbs.MemoryInactive()
        return self._memory_inactive

    @memory_inactive.setter
    def memory_inactive(self, value):
        assert value is None or type(value) == int
        self._memory_inactive = value

    @property
    def memory_percent(self):
        if self._memory_percent is None and self._from_fbs:
            self._memory_percent = self._from_fbs.MemoryPercent()
        return self._memory_percent

    @memory_percent.setter
    def memory_percent(self, value):
        assert value is None or type(value) == float
        self._memory_percent = value

    @property
    def memory_shared(self):
        if self._memory_shared is None and self._from_fbs:
            self._memory_shared = self._from_fbs.MemoryShared()
        return self._memory_shared

    @memory_shared.setter
    def memory_shared(self, value):
        assert value is None or type(value) == int
        self._memory_shared = value

    @property
    def memory_slab(self):
        if self._memory_slab is None and self._from_fbs:
            self._memory_slab = self._from_fbs.MemorySlab()
        return self._memory_slab

    @memory_slab.setter
    def memory_slab(self, value):
        assert value is None or type(value) == int
        self._memory_slab = value

    @property
    def memory_total(self):
        if self._memory_total is None and self._from_fbs:
            self._memory_total = self._from_fbs.MemoryTotal()
        return self._memory_total

    @memory_total.setter
    def memory_total(self, value):
        assert value is None or type(value) == int
        self._memory_total = value

    @property
    def memory_used(self):
        if self._memory_used is None and self._from_fbs:
            self._memory_used = self._from_fbs.MemoryUsed()
        return self._memory_used

    @memory_used.setter
    def memory_used(self, value):
        assert value is None or type(value) == int
        self._memory_used = value

    @property
    def disk_busy_time(self):
        if self._disk_busy_time is None and self._from_fbs:
            self._disk_busy_time = self._from_fbs.DiskBusyTime()
        return self._disk_busy_time

    @disk_busy_time.setter
    def disk_busy_time(self, value):
        assert value is None or type(value) == int
        self._disk_busy_time = value

    @property
    def disk_read_bytes(self):
        if self._disk_read_bytes is None and self._from_fbs:
            self._disk_read_bytes = self._from_fbs.DiskReadBytes()
        return self._disk_read_bytes

    @disk_read_bytes.setter
    def disk_read_bytes(self, value):
        assert value is None or type(value) == int
        self._disk_read_bytes = value

    @property
    def disk_read_count(self):
        if self._disk_read_count is None and self._from_fbs:
            self._disk_read_count = self._from_fbs.DiskReadCount()
        return self._disk_read_count

    @disk_read_count.setter
    def disk_read_count(self, value):
        assert value is None or type(value) == int
        self._disk_read_count = value

    @property
    def disk_read_merged_count(self):
        if self._disk_read_merged_count is None and self._from_fbs:
            self._disk_read_merged_count = self._from_fbs.DiskReadMergedCount()
        return self._disk_read_merged_count

    @disk_read_merged_count.setter
    def disk_read_merged_count(self, value):
        assert value is None or type(value) == int
        self._disk_read_merged_count = value

    @property
    def disk_read_time(self):
        if self._disk_read_time is None and self._from_fbs:
            self._disk_read_time = self._from_fbs.DiskReadTime()
        return self._disk_read_time

    @disk_read_time.setter
    def disk_read_time(self, value):
        assert value is None or type(value) == int
        self._disk_read_time = value

    @property
    def disk_write_bytes(self):
        if self._disk_write_bytes is None and self._from_fbs:
            self._disk_write_bytes = self._from_fbs.DiskWriteBytes()
        return self._disk_write_bytes

    @disk_write_bytes.setter
    def disk_write_bytes(self, value):
        assert value is None or type(value) == int
        self._disk_write_bytes = value

    @property
    def disk_write_count(self):
        if self._disk_write_count is None and self._from_fbs:
            self._disk_write_count = self._from_fbs.DiskWriteCount()
        return self._disk_write_count

    @disk_write_count.setter
    def disk_write_count(self, value):
        assert value is None or type(value) == int
        self._disk_write_count = value

    @property
    def disk_write_merged_count(self):
        if self._disk_write_merged_count is None and self._from_fbs:
            self._disk_write_merged_count = self._from_fbs.DiskWriteMergedCount()
        return self._disk_write_merged_count

    @disk_write_merged_count.setter
    def disk_write_merged_count(self, value):
        assert value is None or type(value) == int
        self._disk_write_merged_count = value

    @property
    def disk_write_time(self):
        if self._disk_write_time is None and self._from_fbs:
            self._disk_write_time = self._from_fbs.DiskWriteTime()
        return self._disk_write_time

    @disk_write_time.setter
    def disk_write_time(self, value):
        assert value is None or type(value) == int
        self._disk_write_time = value

    @staticmethod
    def cast(buf):
        assert type(buf) in [bytes, bytearray], "bytes expected, got {}".format(
            type(buf)
        )
        return MNodeLog(_MNodeLogGen.GetRootAsMNodeLog(buf, 0))

    def build(self, builder):
        node_id = self.node_id.bytes if self.node_id else None
        if node_id:
            node_id = builder.CreateString(node_id)

        run_id = self.run_id.bytes if self.run_id else None
        if run_id:
            run_id = builder.CreateString(run_id)

        MNodeLogGen.MNodeLogStart(builder)

        if self.timestamp:
            MNodeLogGen.MNodeLogAddTimestamp(builder, int(self.timestamp))

        if node_id:
            MNodeLogGen.MNodeLogAddNodeId(builder, node_id)

        if run_id:
            MNodeLogGen.MNodeLogAddRunId(builder, run_id)

        if self.state:
            MNodeLogGen.MNodeLogAddState(builder, self.state)

        if self.ended:
            MNodeLogGen.MNodeLogAddEnded(builder, int(self.ended))

        if self.session:
            MNodeLogGen.MNodeLogAddSession(builder, self.session)

        if self.sent:
            MNodeLogGen.MNodeLogAddSent(builder, int(self.sent))

        if self.seq:
            MNodeLogGen.MNodeLogAddSeq(builder, self.seq)

        if self.routers:
            MNodeLogGen.MNodeLogAddRouters(builder, self.routers)

        if self.containers:
            MNodeLogGen.MNodeLogAddContainers(builder, self.containers)

        if self.guests:
            MNodeLogGen.MNodeLogAddGuests(builder, self.guests)

        if self.proxies:
            MNodeLogGen.MNodeLogAddProxies(builder, self.proxies)

        if self.marketmakers:
            MNodeLogGen.MNodeLogAddMarketmakers(builder, self.marketmakers)

        if self.cpu_ctx_switches:
            MNodeLogGen.MNodeLogAddCpuCtxSwitches(builder, self.cpu_ctx_switches)

        if self.cpu_freq:
            MNodeLogGen.MNodeLogAddCpuFreq(builder, self.cpu_freq)

        if self.cpu_guest:
            MNodeLogGen.MNodeLogAddCpuGuest(builder, self.cpu_guest)

        if self.cpu_guest_nice:
            MNodeLogGen.MNodeLogAddCpuGuestNice(builder, self.cpu_guest_nice)

        if self.cpu_idle:
            MNodeLogGen.MNodeLogAddCpuIdle(builder, self.cpu_idle)

        if self.cpu_interrupts:
            MNodeLogGen.MNodeLogAddCpuInterrupts(builder, self.cpu_interrupts)

        if self.cpu_iotwait:
            MNodeLogGen.MNodeLogAddCpuIowait(builder, self.cpu_iotwait)

        if self.cpu_irq:
            MNodeLogGen.MNodeLogAddCpuIrq(builder, self.cpu_irq)

        if self.cpu_nice:
            MNodeLogGen.MNodeLogAddCpuNice(builder, self.cpu_nice)

        if self.cpu_soft_interrupts:
            MNodeLogGen.MNodeLogAddCpuSoftInterrupts(builder, self.cpu_soft_interrupts)

        if self.cpu_softirq:
            MNodeLogGen.MNodeLogAddCpuSoftirq(builder, self.cpu_softirq)

        if self.cpu_steal:
            MNodeLogGen.MNodeLogAddCpuSteal(builder, self.cpu_steal)

        if self.cpu_system:
            MNodeLogGen.MNodeLogAddCpuSystem(builder, self.cpu_system)

        if self.cpu_user:
            MNodeLogGen.MNodeLogAddCpuUser(builder, self.cpu_user)

        if self.network_bytes_recv:
            MNodeLogGen.MNodeLogAddNetworkBytesRecv(builder, self.network_bytes_recv)

        if self.network_bytes_sent:
            MNodeLogGen.MNodeLogAddNetworkBytesSent(builder, self.network_bytes_sent)

        if self.network_connection_af_inet:
            MNodeLogGen.MNodeLogAddNetworkConnectionAfInet(
                builder, self.network_connection_af_inet
            )

        if self.network_connection_af_inet6:
            MNodeLogGen.MNodeLogAddNetworkConnectionAfInet6(
                builder, self.network_connection_af_inet6
            )

        if self.network_connection_af_unix:
            MNodeLogGen.MNodeLogAddNetworkConnectionAfUnix(
                builder, self.network_connection_af_unix
            )

        if self.network_dropin:
            MNodeLogGen.MNodeLogAddNetworkDropin(builder, self.network_dropin)

        if self.network_dropout:
            MNodeLogGen.MNodeLogAddNetworkDropout(builder, self.network_dropout)

        if self.network_errin:
            MNodeLogGen.MNodeLogAddNetworkErrin(builder, self.network_errin)

        if self.network_errout:
            MNodeLogGen.MNodeLogAddNetworkErrout(builder, self.network_errout)

        if self.network_packets_recv:
            MNodeLogGen.MNodeLogAddNetworkPacketsRecv(
                builder, self.network_packets_recv
            )

        if self.network_packets_sent:
            MNodeLogGen.MNodeLogAddNetworkPacketsSent(
                builder, self.network_packets_sent
            )

        if self.memory_active:
            MNodeLogGen.MNodeLogAddMemoryActive(builder, self.memory_active)

        if self.memory_available:
            MNodeLogGen.MNodeLogAddMemoryAvailable(builder, self.memory_available)

        if self.memory_buffers:
            MNodeLogGen.MNodeLogAddMemoryBuffers(builder, self.memory_buffers)

        if self.memory_cached:
            MNodeLogGen.MNodeLogAddMemoryCached(builder, self.memory_cached)

        if self.memory_free:
            MNodeLogGen.MNodeLogAddMemoryFree(builder, self.memory_free)

        if self.memory_inactive:
            MNodeLogGen.MNodeLogAddMemoryInactive(builder, self.memory_inactive)

        if self.memory_percent:
            MNodeLogGen.MNodeLogAddMemoryPercent(builder, self.memory_percent)

        if self.memory_shared:
            MNodeLogGen.MNodeLogAddMemoryShared(builder, self.memory_shared)

        if self.memory_slab:
            MNodeLogGen.MNodeLogAddMemorySlab(builder, self.memory_slab)

        if self.memory_total:
            MNodeLogGen.MNodeLogAddMemoryTotal(builder, self.memory_total)

        if self.memory_used:
            MNodeLogGen.MNodeLogAddMemoryUsed(builder, self.memory_used)

        if self.disk_busy_time:
            MNodeLogGen.MNodeLogAddDiskBusyTime(builder, self.disk_busy_time)

        if self.disk_read_bytes:
            MNodeLogGen.MNodeLogAddDiskReadBytes(builder, self.disk_read_bytes)

        if self.disk_read_count:
            MNodeLogGen.MNodeLogAddDiskReadCount(builder, self.disk_read_count)

        if self.disk_read_merged_count:
            MNodeLogGen.MNodeLogAddDiskReadMergedCount(
                builder, self.disk_read_merged_count
            )

        if self.disk_read_time:
            MNodeLogGen.MNodeLogAddDiskReadTime(builder, self.disk_read_time)

        if self.disk_write_bytes:
            MNodeLogGen.MNodeLogAddDiskWriteBytes(builder, self.disk_write_bytes)

        if self.disk_write_count:
            MNodeLogGen.MNodeLogAddDiskWriteCount(builder, self.disk_write_count)

        if self.disk_write_merged_count:
            MNodeLogGen.MNodeLogAddDiskWriteMergedCount(
                builder, self.disk_write_merged_count
            )

        if self.disk_write_time:
            MNodeLogGen.MNodeLogAddDiskWriteTime(builder, self.disk_write_time)

        final = MNodeLogGen.MNodeLogEnd(builder)

        return final


@table("256a071f-5aeb-47f3-8786-97cd8281bdb7", build=MNodeLog.build, cast=MNodeLog.cast)
class MNodeLogs(MapTimestampUuidFlatBuffers):
    pass


class Schema(object):
    mnode_logs = None

    def __init__(self, db):
        self.db = db

    @staticmethod
    def attach(db):
        schema = Schema(db)

        schema.mnode_logs = db.attach_table(MNodeLogs)

        return schema



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

import flatbuffers


# /// Logs of runs (from node start to end) of managed CF nodes .
class MNodeLog(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsMNodeLog(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MNodeLog()
        x.Init(buf, n + offset)
        return x

    # MNodeLog
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # /// Unix time in ns when this log record was received (from CFC node clock).
    # MNodeLog

    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # /// CF node ID.
    # MNodeLog

    def NodeId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # MNodeLog
    def NodeIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MNodeLog
    def NodeIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # /// CFC run ID (this is unique over all start-stop cycles of CFC, and constant per run).
    # MNodeLog

    def RunId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # MNodeLog
    def RunIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MNodeLog
    def RunIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # /// Current state of CF node.
    # MNodeLog

    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # /// When the state is ENDED, the end timestamp (Unix time in ns).
    # MNodeLog

    def Ended(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # /// WAMP session ID of the CF node uplink management session to this CFC instance.
    # MNodeLog

    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # /// Unix time in ns. This timestamp is from the original received event payload (from CF node clock).
    # MNodeLog

    def Sent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # /// Sequence number as sent in the log record by the CF node (started at 0 for CF start and incremented by one on each heartbeat).
    # MNodeLog

    def Seq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # /// Number of router workers currently running in the CF node.
    # MNodeLog

    def Routers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # /// Number of container workers currently running in the CF node.
    # MNodeLog

    def Containers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # /// Number of guest workers currently running in the CF node.
    # MNodeLog

    def Guests(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # /// Number of proxy workers currently running in the CF node.
    # MNodeLog

    def Proxies(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # /// Number of XBR market maker workers currently running in the CF node.
    # MNodeLog

    def Marketmakers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # /// CF node system statistics.
    # MNodeLog

    def CpuCtxSwitches(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def CpuFreq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuGuest(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuGuestNice(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuIdle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuInterrupts(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def CpuIowait(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuIrq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuNice(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuSoftInterrupts(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(48))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def CpuSoftirq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(50))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuSteal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(52))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuSystem(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(54))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def CpuUser(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(56))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def DiskBusyTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(58))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def DiskReadBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(60))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def DiskReadCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(62))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def DiskReadMergedCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(64))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def DiskReadTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(66))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def DiskWriteBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(68))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def DiskWriteCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(70))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def DiskWriteMergedCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(72))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def DiskWriteTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(74))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryActive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(76))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryAvailable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(78))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryBuffers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(80))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryCached(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(82))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryFree(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(84))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryInactive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(86))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryPercent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(88))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0

    # MNodeLog
    def MemoryShared(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(90))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemorySlab(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(92))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryTotal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(94))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def MemoryUsed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(96))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkBytesRecv(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(98))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkBytesSent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(100))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkConnectionAfInet(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(102))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkConnectionAfInet6(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(104))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkConnectionAfUnix(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(106))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkDropin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(108))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkDropout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(110))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkErrin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(112))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkErrout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(114))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkPacketsRecv(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(116))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # MNodeLog
    def NetworkPacketsSent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(118))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0


def MNodeLogStart(builder):
    builder.StartObject(58)


def MNodeLogAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def MNodeLogAddNodeId(builder, nodeId):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(nodeId), 0
    )


def MNodeLogStartNodeIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def MNodeLogAddRunId(builder, runId):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(runId), 0
    )


def MNodeLogStartRunIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def MNodeLogAddState(builder, state):
    builder.PrependUint8Slot(3, state, 0)


def MNodeLogAddEnded(builder, ended):
    builder.PrependUint64Slot(4, ended, 0)


def MNodeLogAddSession(builder, session):
    builder.PrependUint64Slot(5, session, 0)


def MNodeLogAddSent(builder, sent):
    builder.PrependUint64Slot(6, sent, 0)


def MNodeLogAddSeq(builder, seq):
    builder.PrependUint64Slot(7, seq, 0)


def MNodeLogAddRouters(builder, routers):
    builder.PrependUint16Slot(8, routers, 0)


def MNodeLogAddContainers(builder, containers):
    builder.PrependUint16Slot(9, containers, 0)


def MNodeLogAddGuests(builder, guests):
    builder.PrependUint16Slot(10, guests, 0)


def MNodeLogAddProxies(builder, proxies):
    builder.PrependUint16Slot(11, proxies, 0)


def MNodeLogAddMarketmakers(builder, marketmakers):
    builder.PrependUint16Slot(12, marketmakers, 0)


def MNodeLogAddCpuCtxSwitches(builder, cpuCtxSwitches):
    builder.PrependUint64Slot(13, cpuCtxSwitches, 0)


def MNodeLogAddCpuFreq(builder, cpuFreq):
    builder.PrependFloat32Slot(14, cpuFreq, 0.0)


def MNodeLogAddCpuGuest(builder, cpuGuest):
    builder.PrependFloat32Slot(15, cpuGuest, 0.0)


def MNodeLogAddCpuGuestNice(builder, cpuGuestNice):
    builder.PrependFloat32Slot(16, cpuGuestNice, 0.0)


def MNodeLogAddCpuIdle(builder, cpuIdle):
    builder.PrependFloat32Slot(17, cpuIdle, 0.0)


def MNodeLogAddCpuInterrupts(builder, cpuInterrupts):
    builder.PrependUint64Slot(18, cpuInterrupts, 0)


def MNodeLogAddCpuIowait(builder, cpuIowait):
    builder.PrependFloat32Slot(19, cpuIowait, 0.0)


def MNodeLogAddCpuIrq(builder, cpuIrq):
    builder.PrependFloat32Slot(20, cpuIrq, 0.0)


def MNodeLogAddCpuNice(builder, cpuNice):
    builder.PrependFloat32Slot(21, cpuNice, 0.0)


def MNodeLogAddCpuSoftInterrupts(builder, cpuSoftInterrupts):
    builder.PrependUint64Slot(22, cpuSoftInterrupts, 0)


def MNodeLogAddCpuSoftirq(builder, cpuSoftirq):
    builder.PrependFloat32Slot(23, cpuSoftirq, 0.0)


def MNodeLogAddCpuSteal(builder, cpuSteal):
    builder.PrependFloat32Slot(24, cpuSteal, 0.0)


def MNodeLogAddCpuSystem(builder, cpuSystem):
    builder.PrependFloat32Slot(25, cpuSystem, 0.0)


def MNodeLogAddCpuUser(builder, cpuUser):
    builder.PrependFloat32Slot(26, cpuUser, 0.0)


def MNodeLogAddDiskBusyTime(builder, diskBusyTime):
    builder.PrependUint64Slot(27, diskBusyTime, 0)


def MNodeLogAddDiskReadBytes(builder, diskReadBytes):
    builder.PrependUint64Slot(28, diskReadBytes, 0)


def MNodeLogAddDiskReadCount(builder, diskReadCount):
    builder.PrependUint64Slot(29, diskReadCount, 0)


def MNodeLogAddDiskReadMergedCount(builder, diskReadMergedCount):
    builder.PrependUint64Slot(30, diskReadMergedCount, 0)


def MNodeLogAddDiskReadTime(builder, diskReadTime):
    builder.PrependUint64Slot(31, diskReadTime, 0)


def MNodeLogAddDiskWriteBytes(builder, diskWriteBytes):
    builder.PrependUint64Slot(32, diskWriteBytes, 0)


def MNodeLogAddDiskWriteCount(builder, diskWriteCount):
    builder.PrependUint64Slot(33, diskWriteCount, 0)


def MNodeLogAddDiskWriteMergedCount(builder, diskWriteMergedCount):
    builder.PrependUint64Slot(34, diskWriteMergedCount, 0)


def MNodeLogAddDiskWriteTime(builder, diskWriteTime):
    builder.PrependUint64Slot(35, diskWriteTime, 0)


def MNodeLogAddMemoryActive(builder, memoryActive):
    builder.PrependUint64Slot(36, memoryActive, 0)


def MNodeLogAddMemoryAvailable(builder, memoryAvailable):
    builder.PrependUint64Slot(37, memoryAvailable, 0)


def MNodeLogAddMemoryBuffers(builder, memoryBuffers):
    builder.PrependUint64Slot(38, memoryBuffers, 0)


def MNodeLogAddMemoryCached(builder, memoryCached):
    builder.PrependUint64Slot(39, memoryCached, 0)


def MNodeLogAddMemoryFree(builder, memoryFree):
    builder.PrependUint64Slot(40, memoryFree, 0)


def MNodeLogAddMemoryInactive(builder, memoryInactive):
    builder.PrependUint64Slot(41, memoryInactive, 0)


def MNodeLogAddMemoryPercent(builder, memoryPercent):
    builder.PrependFloat32Slot(42, memoryPercent, 0.0)


def MNodeLogAddMemoryShared(builder, memoryShared):
    builder.PrependUint64Slot(43, memoryShared, 0)


def MNodeLogAddMemorySlab(builder, memorySlab):
    builder.PrependUint64Slot(44, memorySlab, 0)


def MNodeLogAddMemoryTotal(builder, memoryTotal):
    builder.PrependUint64Slot(45, memoryTotal, 0)


def MNodeLogAddMemoryUsed(builder, memoryUsed):
    builder.PrependUint64Slot(46, memoryUsed, 0)


def MNodeLogAddNetworkBytesRecv(builder, networkBytesRecv):
    builder.PrependUint64Slot(47, networkBytesRecv, 0)


def MNodeLogAddNetworkBytesSent(builder, networkBytesSent):
    builder.PrependUint64Slot(48, networkBytesSent, 0)


def MNodeLogAddNetworkConnectionAfInet(builder, networkConnectionAfInet):
    builder.PrependUint32Slot(49, networkConnectionAfInet, 0)


def MNodeLogAddNetworkConnectionAfInet6(builder, networkConnectionAfInet6):
    builder.PrependUint32Slot(50, networkConnectionAfInet6, 0)


def MNodeLogAddNetworkConnectionAfUnix(builder, networkConnectionAfUnix):
    builder.PrependUint32Slot(51, networkConnectionAfUnix, 0)


def MNodeLogAddNetworkDropin(builder, networkDropin):
    builder.PrependUint32Slot(52, networkDropin, 0)


def MNodeLogAddNetworkDropout(builder, networkDropout):
    builder.PrependUint32Slot(53, networkDropout, 0)


def MNodeLogAddNetworkErrin(builder, networkErrin):
    builder.PrependUint32Slot(54, networkErrin, 0)


def MNodeLogAddNetworkErrout(builder, networkErrout):
    builder.PrependUint32Slot(55, networkErrout, 0)


def MNodeLogAddNetworkPacketsRecv(builder, networkPacketsRecv):
    builder.PrependUint64Slot(56, networkPacketsRecv, 0)


def MNodeLogAddNetworkPacketsSent(builder, networkPacketsSent):
    builder.PrependUint64Slot(57, networkPacketsSent, 0)


def MNodeLogEnd(builder):
    return builder.EndObject()



#!/usr/bin/env python
"""
CFFI builder script for LMDB extension module.

This script uses the modern CFFI API (ffi.set_source) to define the
lmdb_cffi extension module for setuptools integration.

Used by setuptools via cffi_modules configuration in pyproject.toml.
"""

import os
import sys

# Import the configuration generated by build_lmdb.py
# Avoid circular import by importing _config directly
_config_vars = None
try:
    # Try importing _config module directly without triggering zlmdb/lmdb/__init__.py
    import importlib.util
    spec = importlib.util.spec_from_file_location("_config", "zlmdb/lmdb/_config.py")
    if spec and spec.loader:
        _config = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(_config)
        _config_vars = _config.CONFIG
except (FileNotFoundError, ImportError, AttributeError):
    pass

if _config_vars is None:
    # Fallback if _config.py doesn't exist yet
    print("WARNING: zlmdb/lmdb/_config.py not found, using minimal config")
    _config_vars = {
        'extra_compile_args': ['-UNDEBUG', '-DHAVE_PATCHED_LMDB=1', '-w'],
        'extra_sources': ['build/lmdb-src/mdb.c', 'build/lmdb-src/midl.c'],
        'extra_include_dirs': ['build/lmdb-src', 'lmdb-patches'],
        'extra_library_dirs': [],
        'libraries': [],
    }

# Check if we have patched LMDB
_have_patched_lmdb = '-DHAVE_PATCHED_LMDB=1' in _config_vars['extra_compile_args']


# CFFI C definitions (from lmdb/cffi.py)
_CFFI_CDEF = '''
    typedef int mode_t;
    typedef ... MDB_env;
    typedef struct MDB_txn MDB_txn;
    typedef struct MDB_cursor MDB_cursor;
    typedef unsigned int MDB_dbi;

    typedef struct MDB_val {
        size_t mv_size;
        void *mv_data;
    } MDB_val;

    typedef int (*MDB_cmp_func)(const MDB_val *a, const MDB_val *b);
    typedef void (*MDB_rel_func)(MDB_val *item, void *oldptr, void *newptr,
        void *relctx);

    typedef enum MDB_cursor_op {
        MDB_FIRST,
        MDB_FIRST_DUP,
        MDB_GET_BOTH,
        MDB_GET_BOTH_RANGE,
        MDB_GET_CURRENT,
        MDB_GET_MULTIPLE,
        MDB_LAST,
        MDB_LAST_DUP,
        MDB_NEXT,
        MDB_NEXT_DUP,
        MDB_NEXT_MULTIPLE,
        MDB_NEXT_NODUP,
        MDB_PREV,
        MDB_PREV_DUP,
        MDB_PREV_NODUP,
        MDB_SET,
        MDB_SET_KEY,
        MDB_SET_RANGE,
        MDB_PREV_MULTIPLE,
        ...
    } MDB_cursor_op;

    typedef struct MDB_stat {
        unsigned int ms_psize;
        unsigned int ms_depth;
        size_t ms_branch_pages;
        size_t ms_leaf_pages;
        size_t ms_overflow_pages;
        size_t ms_entries;
    } MDB_stat;

    typedef struct MDB_envinfo {
        void *me_mapaddr;
        size_t me_mapsize;
        size_t me_last_pgno;
        size_t me_last_txnid;
        unsigned int me_maxreaders;
        unsigned int me_numreaders;
    } MDB_envinfo;

    char *mdb_version(int *major, int *minor, int *patch);
    char *mdb_strerror(int err);

    int mdb_env_create(MDB_env **env);
    int mdb_env_open(MDB_env *env, const char *path, unsigned int flags,
        mode_t mode);
    int mdb_env_copy2(MDB_env *env, const char *path, unsigned int flags);
    int mdb_env_copyfd2(MDB_env *env, int fd, unsigned int flags);
    int mdb_env_stat(MDB_env *env, MDB_stat *stat);
    int mdb_env_info(MDB_env *env, MDB_envinfo *stat);
    int mdb_env_sync(MDB_env *env, int force);
    void mdb_env_close(MDB_env *env);
    int mdb_env_set_flags(MDB_env *env, unsigned int flags, int onoff);
    int mdb_env_get_flags(MDB_env *env, unsigned int *flags);
    int mdb_env_get_path(MDB_env *env, const char **path);
    int mdb_env_get_fd(MDB_env *env, int *fd);
    int mdb_env_set_mapsize(MDB_env *env, size_t size);
    int mdb_env_set_maxreaders(MDB_env *env, unsigned int readers);
    int mdb_env_get_maxreaders(MDB_env *env, unsigned int *readers);
    int mdb_env_set_maxdbs(MDB_env *env, MDB_dbi dbs);
    int mdb_env_get_maxkeysize(MDB_env *env);
    int mdb_env_set_userctx(MDB_env *env, void *ctx);
    void *mdb_env_get_userctx(MDB_env *env);

    typedef void (*MDB_assert_func)(MDB_env *env, const char *msg);
    int mdb_env_set_assert(MDB_env *env, MDB_assert_func func);

    typedef int (*MDB_msg_func)(const char *msg, void *ctx);

    int mdb_txn_begin(MDB_env *env, MDB_txn *parent, unsigned int flags,
        MDB_txn **txn);
    MDB_env *mdb_txn_env(MDB_txn *txn);
    size_t mdb_txn_id(MDB_txn *txn);
    int mdb_txn_commit(MDB_txn *txn);
    void mdb_txn_abort(MDB_txn *txn);
    void mdb_txn_reset(MDB_txn *txn);
    int mdb_txn_renew(MDB_txn *txn);

    int mdb_dbi_open(MDB_txn *txn, const char *name, unsigned int flags,
        MDB_dbi *dbi);
    int mdb_stat(MDB_txn *txn, MDB_dbi dbi, MDB_stat *stat);
    int mdb_dbi_flags(MDB_txn *txn, MDB_dbi dbi, unsigned int *flags);
    void mdb_dbi_close(MDB_env *env, MDB_dbi dbi);
    int mdb_drop(MDB_txn *txn, MDB_dbi dbi, int del);

    int mdb_set_compare(MDB_txn *txn, MDB_dbi dbi, MDB_cmp_func cmp);
    int mdb_set_dupsort(MDB_txn *txn, MDB_dbi dbi, MDB_cmp_func cmp);
    int mdb_set_relfunc(MDB_txn *txn, MDB_dbi dbi, MDB_rel_func rel);
    int mdb_set_relctx(MDB_txn *txn, MDB_dbi dbi, void *ctx);

    int mdb_get(MDB_txn *txn, MDB_dbi dbi, MDB_val *key, MDB_val *data);
    int mdb_put(MDB_txn *txn, MDB_dbi dbi, MDB_val *key, MDB_val *data,
        unsigned int flags);
    int mdb_del(MDB_txn *txn, MDB_dbi dbi, MDB_val *key, MDB_val *data);

    int mdb_cursor_open(MDB_txn *txn, MDB_dbi dbi, MDB_cursor **cursor);
    void mdb_cursor_close(MDB_cursor *cursor);
    int mdb_cursor_renew(MDB_txn *txn, MDB_cursor *cursor);
    MDB_txn *mdb_cursor_txn(MDB_cursor *cursor);
    MDB_dbi mdb_cursor_dbi(MDB_cursor *cursor);
    int mdb_cursor_get(MDB_cursor *cursor, MDB_val *key, MDB_val *data,
        MDB_cursor_op op);
    int mdb_cursor_put(MDB_cursor *cursor, MDB_val *key, MDB_val *data,
        unsigned int flags);
    int mdb_cursor_del(MDB_cursor *cursor, unsigned int flags);
    int mdb_cursor_count(MDB_cursor *cursor, size_t *countp);

    int mdb_cmp(MDB_txn *txn, MDB_dbi dbi, const MDB_val *a, const MDB_val *b);
    int mdb_dcmp(MDB_txn *txn, MDB_dbi dbi, const MDB_val *a, const MDB_val *b);

    int mdb_reader_list(MDB_env *env, MDB_msg_func func, void *ctx);
    int mdb_reader_check(MDB_env *env, int *dead);

    // LMDB constants for full py-lmdb API compatibility
    // Using ... tells CFFI to extract actual values from lmdb.h automatically

    // Version information
    #define MDB_VERSION_MAJOR ...
    #define MDB_VERSION_MINOR ...
    #define MDB_VERSION_PATCH ...

    // Environment flags (for mdb_env_open)
    #define MDB_FIXEDMAP ...
    #define MDB_NOSUBDIR ...
    #define MDB_RDONLY ...
    #define MDB_WRITEMAP ...
    #define MDB_NOMETASYNC ...
    #define MDB_NOSYNC ...
    #define MDB_MAPASYNC ...
    #define MDB_NOTLS ...
    #define MDB_NOLOCK ...
    #define MDB_NORDAHEAD ...
    #define MDB_NOMEMINIT ...

    // Database flags (for mdb_dbi_open)
    #define MDB_REVERSEKEY ...
    #define MDB_DUPSORT ...
    #define MDB_INTEGERKEY ...
    #define MDB_DUPFIXED ...
    #define MDB_INTEGERDUP ...
    #define MDB_REVERSEDUP ...
    #define MDB_CREATE ...

    // Write flags (for mdb_put)
    #define MDB_NOOVERWRITE ...
    #define MDB_NODUPDATA ...
    #define MDB_CURRENT ...
    #define MDB_RESERVE ...
    #define MDB_APPEND ...
    #define MDB_APPENDDUP ...

    // Copy flags (for mdb_env_copy2)
    #define MDB_CP_COMPACT ...

    // Error codes (LMDB-specific)
    #define MDB_KEYEXIST ...
    #define MDB_NOTFOUND ...
    #define MDB_PAGE_NOTFOUND ...
    #define MDB_CORRUPTED ...
    #define MDB_PANIC ...
    #define MDB_VERSION_MISMATCH ...
    #define MDB_INVALID ...
    #define MDB_MAP_FULL ...
    #define MDB_DBS_FULL ...
    #define MDB_READERS_FULL ...
    #define MDB_TLS_FULL ...
    #define MDB_TXN_FULL ...
    #define MDB_CURSOR_FULL ...
    #define MDB_PAGE_FULL ...
    #define MDB_MAP_RESIZED ...
    #define MDB_INCOMPATIBLE ...
    #define MDB_BAD_RSLOT ...
    #define MDB_BAD_TXN ...
    #define MDB_BAD_VALSIZE ...
    #define MDB_BAD_DBI ...

    // POSIX error codes (from errno.h, used by py-lmdb error handling)
    #define EACCES ...
    #define EAGAIN ...
    #define EINVAL ...
    #define ENOMEM ...
    #define ENOSPC ...

    // py-lmdb helper functions (inline MDB_val construction for performance)
    static int pymdb_del(MDB_txn *txn, MDB_dbi dbi,
                         char *key_s, size_t keylen,
                         char *val_s, size_t vallen);
    static int pymdb_put(MDB_txn *txn, MDB_dbi dbi,
                         char *key_s, size_t keylen,
                         char *val_s, size_t vallen,
                         unsigned int flags);
    static int pymdb_get(MDB_txn *txn, MDB_dbi dbi,
                         char *key_s, size_t keylen,
                         MDB_val *val_out);
    static int pymdb_cursor_get(MDB_cursor *cursor,
                                char *key_s, size_t key_len,
                                char *data_s, size_t data_len,
                                MDB_val *key, MDB_val *data, int op);
    static int pymdb_cursor_put(MDB_cursor *cursor,
                                char *key_s, size_t keylen,
                                char *val_s, size_t vallen, int flags);

    // Memory prefaulting helper
    static void preload(int rc, void *x, size_t size);
'''

# Patched LMDB extensions
_CFFI_CDEF_PATCHED = '''
    int mdb_env_copy3(MDB_env *env, const char *path, unsigned int flags, MDB_txn *txn);
    int mdb_env_copyfd3(MDB_env *env, int fd, unsigned int flags, MDB_txn *txn);
'''

# Combine definitions if we have patched LMDB
if _have_patched_lmdb:
    _CFFI_CDEF += _CFFI_CDEF_PATCHED

# CFFI C code (includes helper function implementations from py-lmdb)
_CFFI_VERIFY = '''
    #include <stdlib.h>
    #include <stdio.h>
    #include <sys/types.h>
    #include <errno.h>
    #include "lmdb.h"
    #include "preload.h"

    // py-lmdb helper functions (inline MDB_val construction for CPython performance)
    static int pymdb_get(MDB_txn *txn, MDB_dbi dbi, char *key_s, size_t keylen,
                         MDB_val *val_out)
    {
        MDB_val key = {keylen, key_s};
        int rc = mdb_get(txn, dbi, &key, val_out);
        return rc;
    }

    static int pymdb_put(MDB_txn *txn, MDB_dbi dbi, char *key_s, size_t keylen,
                         char *val_s, size_t vallen, unsigned int flags)
    {
        MDB_val key = {keylen, key_s};
        MDB_val val = {vallen, val_s};
        return mdb_put(txn, dbi, &key, &val, flags);
    }

    static int pymdb_del(MDB_txn *txn, MDB_dbi dbi, char *key_s, size_t keylen,
                         char *val_s, size_t vallen)
    {
        MDB_val key = {keylen, key_s};
        MDB_val val = {vallen, val_s};
        MDB_val *valptr;
        if(vallen == 0) {
            valptr = NULL;
        } else {
            valptr = &val;
        }
        return mdb_del(txn, dbi, &key, valptr);
    }

    static int pymdb_cursor_get(MDB_cursor *cursor,
                                char *key_s, size_t key_len,
                                char *data_s, size_t data_len,
                                MDB_val *key, MDB_val *data, int op)
    {
        MDB_val tmp_key = {key_len, key_s};
        MDB_val tmp_data = {data_len, data_s};
        int rc = mdb_cursor_get(cursor, &tmp_key, &tmp_data, op);
        if(! rc) {
            *key = tmp_key;
            *data = tmp_data;
        }
        return rc;
    }

    static int pymdb_cursor_put(MDB_cursor *cursor, char *key_s, size_t keylen,
                                char *val_s, size_t vallen, int flags)
    {
        MDB_val tmpkey = {keylen, key_s};
        MDB_val tmpval = {vallen, val_s};
        return mdb_cursor_put(cursor, &tmpkey, &tmpval, flags);
    }
'''


# Create FFI instance
import cffi
ffi = cffi.FFI()
ffi.cdef(_CFFI_CDEF)

# Set source using modern CFFI API
ffi.set_source(
    "zlmdb.lmdb._lmdb_cffi",
    _CFFI_VERIFY,
    sources=_config_vars['extra_sources'],
    include_dirs=_config_vars['extra_include_dirs'],
    extra_compile_args=_config_vars['extra_compile_args'],
    libraries=_config_vars['libraries'],
    library_dirs=_config_vars['extra_library_dirs'],
)

# This script exports 'ffi' for setuptools cffi_modules
if __name__ == '__main__':
    ffi.compile(verbose=True)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import random
import os
import zlmdb.lmdb as lmdb
import struct
import platform

import pytest

try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory  # type:ignore


def test_lmdb_create():
    """
    Test creation of LMDB database.
    """
    with TemporaryDirectory() as dbpath:
        env = lmdb.open(dbpath)

        with env.begin() as txn:
            assert txn.id() == 0


def test_lmdb_insert_empty_key_raises():
    """
    Test that LMDB raises on inserting record with empty (bytes) key.
    """
    with TemporaryDirectory() as dbpath:
        env = lmdb.open(dbpath)

        with env.begin(write=True) as txn:
            key = b""
            value = random.randint(0, 2**32 - 1)
            data = struct.pack("<L", value)
            with pytest.raises(lmdb.BadValsizeError):
                txn.put(key, data)


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="behavioral difference on pypy (does not raise, but runs into traceback)",
)
def test_lmdb_insert_null_key_raises():
    """
    Test that LMDB raises on inserting record with NULL key.
    """
    with TemporaryDirectory() as dbpath:
        env = lmdb.open(dbpath)

        with env.begin(write=True) as txn:
            key = None
            value = random.randint(0, 2**32 - 1)
            data = struct.pack("<L", value)
            with pytest.raises(lmdb.BadValsizeError):
                txn.put(key, data)


def test_lmdb_insert_empty_value_ok():
    """
    Test that LMDB allows to insert record with empty (bytes) value.
    """
    with TemporaryDirectory() as dbpath:
        env = lmdb.open(dbpath)

        with env.begin(write=True) as txn:
            key = b"foo"
            data = b""
            txn.put(key, data)

        with env.begin() as txn:
            cursor = txn.cursor()
            assert cursor.first()
            assert cursor.value() == b""
            total = 1
            while cursor.next():
                total += 1
            assert total == 1


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="behavioral difference on pypy (does not raise, but runs into traceback)",
)
def test_lmdb_insert_null_value_ok():
    """
    Test that LMDB allows to insert record with NULL value and
    returns b'' value when retrieving the record (!).
    """
    with TemporaryDirectory() as dbpath:
        env = lmdb.open(dbpath)

        with env.begin(write=True) as txn:
            key = b"foo"
            data = None
            txn.put(key, data)

        with env.begin() as txn:
            cursor = txn.cursor()
            assert cursor.first()
            assert cursor.value() == b""
            total = 1
            while cursor.next():
                total += 1
            assert total == 1


def test_lmdb_insert():
    """
    Test inserting and retrieving a couple of records.
    """
    with TemporaryDirectory() as dbpath:
        n = 100
        total1 = 0
        env = lmdb.open(dbpath)

        with env.begin(write=True) as txn:
            for i in range(n):
                key = "key-{}".format(i).encode("utf8")
                value = random.randint(0, 2**32 - 1)
                total1 += value
                data = struct.pack("<L", value)
                txn.put(key, data)

        with env.begin() as txn:
            cursor = txn.cursor()
            assert cursor.first()

            count = 1
            total2 = struct.unpack("<L", cursor.value())[0]

            while cursor.next():
                count += 1
                total2 += struct.unpack("<L", cursor.value())[0]

            assert count == n
            assert total1 == total2


def test_lmdb_delete():
    """
    Test inserting and deleting again a couple of records.
    """
    with TemporaryDirectory() as dbpath:
        n = 100
        env = lmdb.open(dbpath)

        with env.begin(write=True) as txn:
            for i in range(n):
                key = "key-{}".format(i).encode("utf8")
                value = os.urandom(16)
                txn.put(key, value)

        with env.begin(write=True) as txn:
            for i in range(n):
                key = "key-{}".format(i).encode("utf8")
                assert txn.delete(key)

        with env.begin() as txn:
            cursor = txn.cursor()
            assert not cursor.first()



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os
import sys
import random
import logging

import txaio

txaio.use_twisted()

try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory  # type:ignore

import zlmdb  # noqa

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from _schema_fbs import User  # noqa


class UsersSchema(zlmdb.Schema):
    def __init__(self):
        self.tab_oid_fbs = zlmdb.MapOidFlatBuffers(1, build=User.build, cast=User.cast)


def test_pmap_flatbuffers_values():
    with TemporaryDirectory() as dbpath:
        logging.info("Using temporary directory {} for database".format(dbpath))

        schema = UsersSchema()

        with zlmdb.Database(dbpath) as db:
            N = 100
            stats = zlmdb.TransactionStats()

            with db.begin(write=True, stats=stats) as txn:
                for i in range(N):
                    user = User.create_test_user()
                    schema.tab_oid_fbs[txn, user.oid] = user

            assert stats.puts == N
            assert stats.dels == 0
            stats.reset()

            with db.begin() as txn:
                cnt = schema.tab_oid_fbs.count(txn)

            assert cnt == N


def test_pmap_flatbuffers_count():
    with TemporaryDirectory() as dbpath:
        logging.info("Using temporary directory {} for database".format(dbpath))

        schema = UsersSchema()

        # max DB size is 100 MB
        with zlmdb.Database(dbpath, maxsize=100 * (2**20)) as db:
            oids = set()
            oid_to_referred_by = {}

            stats = zlmdb.TransactionStats()

            # number of transactions
            M = 5

            # number of insert rows per transaction
            N = 10000
            for j in range(M):
                with db.begin(write=True, stats=stats) as txn:
                    for i in range(N):
                        user = User.create_test_user()
                        schema.tab_oid_fbs[txn, user.oid] = user
                        oids.add(user.oid)
                        oid_to_referred_by[user.oid] = user.referred_by

                assert stats.puts == N
                assert stats.dels == 0
                duration_ns = stats.duration
                duration_ms = int(duration_ns / 1000000.0)
                rows_per_sec = int(
                    round(float(stats.puts + stats.dels) * 1000.0 / float(duration_ms))
                )
                logging.info(
                    "Transaction ended: puts={} / dels={} rows in {} ms, {} rows/sec".format(
                        stats.puts, stats.dels, duration_ms, rows_per_sec
                    )
                )

                stats.reset()

            # count all rows
            with db.begin() as txn:
                cnt = schema.tab_oid_fbs.count(txn)

            assert cnt == N * M

            # retrieve
            with db.begin() as txn:
                for j in range(5):
                    started = zlmdb.walltime()
                    M = 100
                    for i in range(M):
                        for oid in random.sample(oids, N):
                            user = schema.tab_oid_fbs[txn, oid]
                            assert user
                            assert user.referred_by == oid_to_referred_by.get(oid, None)
                    duration_ns = zlmdb.walltime() - started
                    duration_ms = int(duration_ns / 1000000.0)
                    rows_per_sec = int(
                        round(float(M * N) * 1000.0 / float(duration_ms))
                    )
                    logging.info(
                        "Transaction ended: {} rows read in {} ms, {} rows/sec".format(
                            M * N, duration_ms, rows_per_sec
                        )
                    )



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import random
import uuid
import datetime

from zlmdb.flatbuffers.demo import User as _user
from zlmdb.flatbuffers.demo import Date as _date


class User(object):
    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._name = None
        self._authid = None
        self._uuid = None
        self._email = None
        self._birthday = None
        self._is_friendly = None
        self._tags = None
        self._ratings = None
        self._ratings_cached = None
        self._friends = None
        self._friends_cached = None
        self._referred_by = None

    @property
    def name(self):
        return self._name or self._from_fbs.Name()

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def authid(self):
        return self._authid or self._from_fbs.Authid()

    @authid.setter
    def authid(self, value):
        self._authid = value

    @property
    def uuid(self):
        return self._uuid or self._from_fbs.Uuid()

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def email(self):
        return self._email or self._from_fbs.Email()

    @email.setter
    def email(self, value):
        self._email = value

    @property
    def birthday(self):
        return self._birthday or self._from_fbs.Birthday()

    @birthday.setter
    def birthday(self, value):
        self._birthday = value

    @property
    def is_friendly(self):
        return self._is_friendly or self._from_fbs.IsFriendly()

    @is_friendly.setter
    def is_friendly(self, value):
        self._is_friendly = value

    @property
    def ratings(self):
        if self._ratings is not None:
            return self._ratings
        if self._ratings_cached is None:
            self._ratings_cached = {}
            if self._from_fbs:
                for i in range(self._from_fbs.RatingsLength()):
                    rat = self._from_fbs.Ratings(i)
                    self._ratings_cached[rat.Name()] = rat.Rating()
        return self._ratings_cached

    @ratings.setter
    def ratings(self, value):
        self._ratings = value

    @property
    def friends(self):
        if self._friends is not None:
            return self._friends
        if self._friends_cached is None:
            self._friends_cached = []
            if self._from_fbs:
                for i in range(self._from_fbs.FriendsLength()):
                    friend_oid = self._from_fbs.Friends(i)
                    self._friends_cached.append(friend_oid)
        return self._friends_cached

    @friends.setter
    def friends(self, value):
        self._friends = value

    @property
    def referred_by(self):
        return self._referred_by or self._from_fbs.ReferredBy()

    @referred_by.setter
    def referred_by(self, value):
        self._referred_by = value

    def build(self, builder):
        if self._name is not None:
            name = builder.CreateString(self._name)
        else:
            name = builder.CreateString(self._from_fbs.Name())

        if self._authid is not None:
            authid = builder.CreateString(self._authid)
        else:
            authid = builder.CreateString(self._from_fbs.Authid())

        if self._email is not None:
            email = builder.CreateString(self._email)
        else:
            email = builder.CreateString(self._from_fbs.Email())

        _user.UserStart(builder)
        _user.UserAddName(builder, name)
        _user.UserAddAuthid(builder, authid)
        _user.UserAddEmail(builder, email)

        if self._birthday is not None:
            _user.UserAddBirthday(
                builder,
                _date.CreateDate(
                    builder,
                    self._birthday.year,
                    self._birthday.month,
                    self._birthday.day,
                ),
            )
        else:
            bd = self._from_fbs.Birthday()
            _user.UserAddBirthday(
                builder, _date.CreateDate(builder, bd.Year(), bd.Month(), bd.Day())
            )

        # FIXME: tags
        # FIXME: ratings
        # FIXME: friends

        if self._is_friendly is not None:
            _user.UserAddIsFriendly(builder, self._is_friendly)
        else:
            _user.UserAddIsFriendly(builder, self._from_fbs.IsFriendly())

        if self._referred_by is not None:
            _user.UserAddReferredBy(builder, self._referred_by)
        else:
            _user.UserAddReferredBy(builder, self._from_fbs.ReferredBy())

        return _user.UserEnd(builder)

    @staticmethod
    def cast(buf):
        return User(_user.User.GetRootAsUser(buf, 0))

    @staticmethod
    def create_test_user(oid=None):
        user = User()
        if oid is not None:
            user.oid = oid
        else:
            user.oid = random.randint(0, 9007199254740992)
        user.name = "Test {}".format(user.oid)
        user.authid = "test-{}".format(user.oid)
        user.uuid = uuid.uuid4()
        user.email = "{}@example.com".format(user.authid)
        user.birthday = datetime.date(1950, 12, 24)
        user.is_friendly = True
        user.tags = ["geek", "sudoko", "yellow"]
        for j in range(10):
            user.ratings["test-rating-{}".format(j)] = random.random()
        user.friends = [random.randint(0, 9007199254740992) for _ in range(10)]
        user.referred_by = random.randint(0, 9007199254740992)
        return user



#
# Copyright 2013-2021 The py-lmdb authors, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted only as authorized by the OpenLDAP
# Public License.
#
# A copy of this license is available in the file LICENSE in the
# top-level directory of the distribution or, alternatively, at
# <http://www.OpenLDAP.org/license.html>.
#
# OpenLDAP is a registered trademark of the OpenLDAP Foundation.
#
# Individual files and/or contributed packages may be copyright by
# other parties and/or subject to additional restrictions.
#
# This work also contains materials derived from public sources.
#
# Additional information about OpenLDAP can be obtained at
# <http://www.openldap.org/>.
#

# test delete(dupdata)

from __future__ import absolute_import
from __future__ import with_statement
import sys
import unittest

import zlmdb.lmdb as lmdb

from . import testlib
from .testlib import B
from .testlib import BT


class ContextManagerTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def test_ok(self):
        path, env = testlib.temp_env()
        txn = env.begin(write=True)
        with txn.cursor() as curs:
            curs.put(B("foo"), B("123"))
        self.assertRaises(Exception, lambda: curs.get(B("foo")))

    def test_crash(self):
        path, env = testlib.temp_env()
        txn = env.begin(write=True)

        try:
            with txn.cursor() as curs:
                curs.put(123, 123)
        except Exception:
            pass
        self.assertRaises(Exception, lambda: curs.get(B("foo")))


class CursorTestBase(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def setUp(self):
        self.path, self.env = testlib.temp_env()
        self.txn = self.env.begin(write=True)
        self.c = self.txn.cursor()


class CursorTest(CursorTestBase):
    def testKeyValueItemEmpty(self):
        self.assertEqual(B(""), self.c.key())
        self.assertEqual(B(""), self.c.value())
        self.assertEqual(BT("", ""), self.c.item())

    def testFirstLastEmpty(self):
        self.assertEqual(False, self.c.first())
        self.assertEqual(False, self.c.last())

    def testFirstFilled(self):
        testlib.putData(self.txn)
        self.assertEqual(True, self.c.first())
        self.assertEqual(testlib.ITEMS[0], self.c.item())

    def testLastFilled(self):
        testlib.putData(self.txn)
        self.assertEqual(True, self.c.last())
        self.assertEqual(testlib.ITEMS[-1], self.c.item())

    def testSetKey(self):
        self.assertRaises(Exception, (lambda: self.c.set_key(B(""))))
        self.assertEqual(False, self.c.set_key(B("missing")))
        testlib.putData(self.txn)
        self.assertEqual(True, self.c.set_key(B("b")))
        self.assertEqual(False, self.c.set_key(B("ba")))

    def testSetRange(self):
        self.assertEqual(False, self.c.set_range(B("x")))
        testlib.putData(self.txn)
        self.assertEqual(False, self.c.set_range(B("x")))
        self.assertEqual(True, self.c.set_range(B("a")))
        self.assertEqual(B("a"), self.c.key())
        self.assertEqual(True, self.c.set_range(B("ba")))
        self.assertEqual(B("baa"), self.c.key())
        self.c.set_range(B(""))
        self.assertEqual(B("a"), self.c.key())

    def testDeleteEmpty(self):
        self.assertEqual(False, self.c.delete())

    def testDeleteFirst(self):
        testlib.putData(self.txn)
        self.assertEqual(False, self.c.delete())
        self.c.first()
        self.assertEqual(BT("a", ""), self.c.item())
        self.assertEqual(True, self.c.delete())
        self.assertEqual(BT("b", ""), self.c.item())
        self.assertEqual(True, self.c.delete())
        self.assertEqual(BT("baa", ""), self.c.item())
        self.assertEqual(True, self.c.delete())
        self.assertEqual(BT("d", ""), self.c.item())
        self.assertEqual(True, self.c.delete())
        self.assertEqual(BT("", ""), self.c.item())
        self.assertEqual(False, self.c.delete())
        self.assertEqual(BT("", ""), self.c.item())

    def testDeleteLast(self):
        testlib.putData(self.txn)
        self.assertEqual(True, self.c.last())
        self.assertEqual(BT("d", ""), self.c.item())
        self.assertEqual(True, self.c.delete())
        self.assertEqual(BT("", ""), self.c.item())
        self.assertEqual(False, self.c.delete())
        self.assertEqual(BT("", ""), self.c.item())

    def testCount(self):
        self.assertRaises(Exception, (lambda: self.c.count()))
        testlib.putData(self.txn)
        self.c.first()
        # TODO: complete dup key support.
        # self.assertEqual(1, self.c.count())

    def testPut(self):
        pass


class CursorTest2(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def setUp(self):
        self.path, self.env = testlib.temp_env()
        self.db = self.env.open_db(b"foo", dupsort=True)
        self.txn = self.env.begin(write=True, db=self.db)
        self.c = self.txn.cursor()

    def testIterWithDeletes(self):
        """A problem identified in LMDB 0.9.27"""
        self.c.put(b"\x00\x01", b"hehe", dupdata=True)
        self.c.put(b"\x00\x02", b"haha", dupdata=True)
        self.c.set_key(b"\x00\x02")
        it = self.c.iternext()
        self.assertEqual((b"\x00\x02", b"haha"), next(it))
        self.txn.delete(b"\x00\x01", b"hehe", db=self.db)
        self.assertRaises(StopIteration, next, it)


class PutmultiTest(CursorTestBase):
    def test_empty_seq(self):
        consumed, added = self.c.putmulti(())
        assert consumed == added == 0

    def test_2list(self):
        l = [BT("a", ""), BT("a", "")]
        consumed, added = self.c.putmulti(l)
        assert consumed == added == 2

        li = iter(l)
        consumed, added = self.c.putmulti(li)
        assert consumed == added == 2

    def test_2list_preserve(self):
        l = [BT("a", ""), BT("a", "")]
        consumed, added = self.c.putmulti(l, overwrite=False)
        assert consumed == 2
        assert added == 1

        assert self.c.set_key(B("a"))
        assert self.c.delete()

        li = iter(l)
        consumed, added = self.c.putmulti(li, overwrite=False)
        assert consumed == 2
        assert added == 1

    def test_bad_seq1(self):
        self.assertRaises(Exception, lambda: self.c.putmulti(range(2)))

    def test_dupsort(self):
        _, env = testlib.temp_env()
        db1 = env.open_db(B("db1"), dupsort=True)
        txn = env.begin(write=True, db=db1)
        with txn.cursor() as c:
            tups = [BT("a", "value1"), BT("b", "value1"), BT("b", "value2")]
            assert (3, 3) == c.putmulti(tups)

    def test_dupsort_putmulti_append(self):
        _, env = testlib.temp_env()
        db1 = env.open_db(B("db1"), dupsort=True)
        txn = env.begin(write=True, db=db1)
        with txn.cursor() as c:
            tups = [BT("a", "value1"), BT("b", "value1"), BT("b", "value2")]
            assert (3, 3) == c.putmulti(tups, append=True)

    def test_dupsort_put_append(self):
        _, env = testlib.temp_env()
        db1 = env.open_db(B("db1"), dupsort=True)
        txn = env.begin(write=True, db=db1)
        with txn.cursor() as c:
            assert c.put(B("a"), B("value1"), append=True)
            assert c.put(B("b"), B("value1"), append=True)
            assert c.put(B("b"), B("value2"), append=True)


class ReplaceTest(CursorTestBase):
    def test_replace(self):
        assert None is self.c.replace(B("a"), B(""))
        assert B("") == self.c.replace(B("a"), B("x"))
        assert B("x") == self.c.replace(B("a"), B("y"))


class ContextManagerTest2(CursorTestBase):
    def test_enter(self):
        with self.c as c:
            assert c is self.c
            c.put(B("a"), B("a"))
            assert c.get(B("a")) == B("a")
        self.assertRaises(Exception, lambda: c.get(B("a")))

    def test_exit_success(self):
        with self.txn.cursor() as c:
            c.put(B("a"), B("a"))
        self.assertRaises(Exception, lambda: c.get(B("a")))

    def test_exit_failure(self):
        try:
            with self.txn.cursor() as c:
                c.put(B("a"), B("a"))
            raise ValueError
        except ValueError:
            pass
        self.assertRaises(Exception, lambda: c.get(B("a")))

    def test_close(self):
        self.c.close()
        self.assertRaises(Exception, lambda: c.get(B("a")))

    def test_double_close(self):
        self.c.close()
        self.c.close()
        self.assertRaises(Exception, lambda: self.c.put(B("a"), B("a")))


GiB = 1024 * 1024 * 1024


class PreloadTest(CursorTestBase):
    def setUp(self, redo=False):
        env_args = {"writemap": True, "map_size": GiB}
        if not redo:
            self.path, self.env = testlib.temp_env(**env_args)
        else:
            self.path, self.env = testlib.temp_env(path=self.path, **env_args)
        self.txn = self.env.begin(write=True)
        self.c = self.txn.cursor()

    @unittest.skipIf(not sys.platform.startswith("linux"), "test only works on Linux")
    def test_preload(self):
        """
        Test that reading just the key doesn't prefault the value contents, but
        reading the data does.
        """

        import resource

        self.c.put(B("a"), B("a") * (256 * 1024 * 1024))
        self.txn.commit()
        self.env.close()
        # Just reading the data is obviously going to fault the value in.  The
        # point is to fault it in while the GIL is unlocked.  We use the buffers
        # API so that we're not actually copying the data in.  This doesn't
        # actually show that we're prefaulting with the GIL unlocked, but it
        # does prove we prefault at all, and in 2 correct places.
        self.path, self.env = testlib.temp_env(path=self.path, writemap=True)
        self.txn = self.env.begin(write=True, buffers=True)
        self.c = self.txn.cursor()
        minflts_before = resource.getrusage(resource.RUSAGE_SELF)[6]
        self.c.set_key(B("a"))
        assert bytes(self.c.key()) == B("a")
        minflts_after_key = resource.getrusage(resource.RUSAGE_SELF)[6]

        self.c.value()
        minflts_after_value = resource.getrusage(resource.RUSAGE_SELF)[6]

        epsilon = 60

        # Setting the position doesn't prefault the data
        assert minflts_after_key - minflts_before < epsilon

        # Getting the value does prefault the data, even if we only get it by pointer
        assert minflts_after_value - minflts_after_key > 1000


class CursorReadOnlyTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def test_cursor_readonly(self):
        """
        Tests whether you can open a cursor on a sub-db at all in a read-only environment.
        """
        path, env = testlib.temp_env(max_dbs=10)
        env.open_db(b"foo")
        env.close()
        with lmdb.open(path, max_dbs=10, readonly=True) as env:
            db2 = env.open_db(b"foo")
            with env.begin(db=db2) as txn:
                with txn.cursor(db=db2):
                    pass


if __name__ == "__main__":
    unittest.main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: reflection

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Service(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Service()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsService(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ServiceBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x42\x46\x42\x53", size_prefixed=size_prefixed)

    # Service
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Service
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Service
    def Calls(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from reflection.RPCCall import RPCCall
            obj = RPCCall()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Service
    def CallsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Service
    def CallsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Service
    def Attributes(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from reflection.KeyValue import KeyValue
            obj = KeyValue()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Service
    def AttributesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Service
    def AttributesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Service
    def Documentation(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Service
    def DocumentationLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Service
    def DocumentationIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # File that this Service is declared in.
    # Service
    def DeclarationFile(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def ServiceStart(builder): builder.StartObject(5)
def Start(builder):
    return ServiceStart(builder)
def ServiceAddName(builder, name): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def AddName(builder, name):
    return ServiceAddName(builder, name)
def ServiceAddCalls(builder, calls): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(calls), 0)
def AddCalls(builder, calls):
    return ServiceAddCalls(builder, calls)
def ServiceStartCallsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartCallsVector(builder, numElems):
    return ServiceStartCallsVector(builder, numElems)
def ServiceAddAttributes(builder, attributes): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(attributes), 0)
def AddAttributes(builder, attributes):
    return ServiceAddAttributes(builder, attributes)
def ServiceStartAttributesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartAttributesVector(builder, numElems):
    return ServiceStartAttributesVector(builder, numElems)
def ServiceAddDocumentation(builder, documentation): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(documentation), 0)
def AddDocumentation(builder, documentation):
    return ServiceAddDocumentation(builder, documentation)
def ServiceStartDocumentationVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartDocumentationVector(builder, numElems):
    return ServiceStartDocumentationVector(builder, numElems)
def ServiceAddDeclarationFile(builder, declarationFile): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(declarationFile), 0)
def AddDeclarationFile(builder, declarationFile):
    return ServiceAddDeclarationFile(builder, declarationFile)
def ServiceEnd(builder): return builder.EndObject()
def End(builder):
    return ServiceEnd(builder)


###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os
import sys
import timeit
import uuid
import platform
import logging

import humanize

import txaio

txaio.use_twisted()

from zlmdb import _types  # noqa
from _schema_fbs import User as UserFbs  # noqa

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

if sys.version_info >= (3, 6):
    from _schema_py3 import User
else:
    from _schema_py2 import User

_TEST = {"oid": 0, "uuid": None, "bytes": 0, "serializer": None}


def _serializer_run_fbs():
    serialize = _TEST["serializer"]._serialize_value
    user = UserFbs.create_test_user()
    data = serialize(user)
    _TEST["bytes"] += len(data)


def _serializer_run():
    serialize = _TEST["serializer"]._serialize_value
    user = User.create_test_user()
    data = serialize(user)
    _TEST["bytes"] += len(data)


def _serialization_speed(serializer, testfun):
    N = 10
    M = 10000

    samples = []

    logging.info("running on:")
    logging.info(sys.version)
    logging.info(platform.uname())

    _TEST["oid"] = 0
    _TEST["uuid"] = uuid.uuid4()
    _TEST["bytes"] = 0
    _TEST["bytes"] = 0
    _TEST["serializer"] = serializer

    for i in range(N):
        secs = timeit.timeit(testfun, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        logging.info(
            "{} objects/sec {}".format(ops, humanize.naturalsize(_TEST["bytes"]))
        )

    ops_max = max(samples)
    bytes_per_obj = float(_TEST["bytes"]) / float(N * M)
    logging.info(
        "{} objects/sec max, {} bytes total, {} bytes/obj".format(
            ops_max,
            humanize.naturalsize(_TEST["bytes"]),
            humanize.naturalsize(bytes_per_obj),
        )
    )

    return ops_max, _TEST["bytes"]


def test_json_serialization_speed():
    ser = _types._JsonValuesMixin(marshal=User.marshal, unmarshal=User.parse)
    ops_max, total = _serialization_speed(ser, _serializer_run)
    # cpy36: 19564.6 objects/sec max, 135456153 bytes total
    assert ops_max > 1000
    assert total > 1000000


def test_cbor_serialization_speed():
    ser = _types._CborValuesMixin(marshal=User.marshal, unmarshal=User.parse)
    ops_max, total = _serialization_speed(ser, _serializer_run)
    # cpy36: 7787.4 objects/sec max, 97815364 bytes total
    assert ops_max > 1000
    assert total > 1000000


def test_pickle_serialization_speed():
    ser = _types._PickleValuesMixin()
    ops_max, total = _serialization_speed(ser, _serializer_run)
    # cpy36: 33586.0 objects/sec max, 137738869 bytes total
    assert ops_max > 1000
    assert total > 1000000


def test_flatbuffer_serialization_speed():
    ser = _types._FlatBuffersValuesMixin(build=UserFbs.build, cast=UserFbs.cast)
    ops_max, total = _serialization_speed(ser, _serializer_run_fbs)
    assert ops_max > 1000
    assert total > 1000000


if __name__ == "__main__":
    from typing import List

    sers: List[object] = []
    sers.append(_types._JsonValuesMixin(marshal=User.marshal, unmarshal=User.parse))
    sers.append(_types._CborValuesMixin(marshal=User.marshal, unmarshal=User.parse))
    sers.append(_types._PickleValuesMixin())
    sers.append(_types._FlatBuffersValuesMixin(build=UserFbs.build, cast=UserFbs.cast))
    for ser in sers:
        logging.info(_serialization_speed(ser, _serializer_run))



#!/usr/bin/env python
"""
Minimal setup.py shim for zlmdb.

This file exists to:
1. Run build_lmdb.py to prepare LMDB sources before building
2. Build the CFFI extension module for LMDB
3. Ensure compatibility with tools that expect setup.py
4. Delegate to pyproject.toml for all modern packaging configuration

For modern builds, use:
    python -m build

For editable installs:
    pip install -e .
"""

import sys
import os
import subprocess
import shutil
import glob
from setuptools import setup
from setuptools.command.build_py import build_py as _build_py
from setuptools.dist import Distribution


# Run build_lmdb.py at module import time to prepare LMDB sources
print("=" * 70)
print("zlmdb setup.py - Preparing LMDB sources")
print("=" * 70)
result = subprocess.run([sys.executable, 'build_lmdb.py'], capture_output=False)
if result.returncode != 0:
    print("\nERROR: build_lmdb.py failed!")
    sys.exit(1)

# Compile CFFI extension at module-level (before setup() call)
# This ensures the .so file exists when setuptools packages the wheel
# Always compile for the current Python interpreter (don't reuse cached .so from other versions)
print("\n" + "=" * 70)
print("zlmdb setup.py - Compiling CFFI extension")
print("=" * 70)
print(f"Building CFFI extension for {sys.version.split()[0]}...")
result = subprocess.run([sys.executable, 'build_cffi_lmdb.py'], capture_output=False)
if result.returncode != 0:
    print("\nERROR: build_cffi_lmdb.py failed!")
    sys.exit(1)
print("OK: CFFI extension compiled successfully")

# Clean up old .so files from other Python versions to prevent cross-contamination
import sysconfig
ext_suffix = sysconfig.get_config_var('EXT_SUFFIX') or sysconfig.get_config_var('SO')
current_so = f'zlmdb/lmdb/_lmdb_cffi{ext_suffix}'
print(f"\nCleaning up .so files from other Python versions (keeping {os.path.basename(current_so)})...")
for so_file in glob.glob('zlmdb/lmdb/_lmdb_cffi*.so') + glob.glob('zlmdb/lmdb/_lmdb_cffi*.pyd'):
    if so_file != current_so:
        print(f"  Removing: {os.path.basename(so_file)}")
        os.remove(so_file)
print("=" * 70 + "\n")


class BuildPyWithCFFI(_build_py):
    """Custom build_py that compiles CFFI extension and copies it to build dir."""

    def run(self):
        # Run standard build_py first
        print("\nRunning standard build_py...")
        _build_py.run(self)

        # Copy compiled CFFI extension for CURRENT Python version to build directory
        print("\nLooking for CFFI extension for current Python version...")

        # Get the expected extension filename for the current Python
        import sysconfig
        ext_suffix = sysconfig.get_config_var('EXT_SUFFIX') or sysconfig.get_config_var('SO')
        expected_file = f'zlmdb/lmdb/_lmdb_cffi{ext_suffix}'

        if os.path.exists(expected_file):
            dest_dir = os.path.join(self.build_lib, 'zlmdb', 'lmdb')
            dest_file = os.path.join(dest_dir, os.path.basename(expected_file))
            print(f"OK: Copying: {os.path.basename(expected_file)} -> {dest_file}")
            shutil.copy2(expected_file, dest_file)
        else:
            print(f"WARNING: Expected CFFI extension not found: {expected_file}")
        print("=" * 70 + "\n")


class BinaryDistribution(Distribution):
    """Force setuptools to recognize this as a binary (platform-specific) package.

    This ensures wheels are tagged correctly (e.g., cp311-cp311-linux_x86_64)
    instead of pure Python tags (py3-none-any).
    """
    def has_ext_modules(self):
        return True


# Configure setuptools with custom build command
setup(
    cmdclass={
        'build_py': BuildPyWithCFFI,
    },
    distclass=BinaryDistribution,
)



# Copyright (c) FlatBuffers Contributors, Apache License 2.0

# Copied from (master branch, 05/29/2022):
# * https://github.com/google/flatbuffers/blob/master/reflection/reflection.fbs
# * https://github.com/google/flatbuffers/tree/master/python/flatbuffers/reflection
